# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Executes queries in BigQuery."""

from __future__ import annotations

import contextlib
import os

try:
  from google.cloud import bigquery  # type: ignore
except ImportError as e:
  raise ImportError(
    'Please install garf-executors with BigQuery support '
    '- `pip install garf-executors[bq]`'
  ) from e

import logging

from garf.core import query_editor, report
from garf.executors import exceptions, execution_context, executor
from garf.executors.telemetry import tracer
from google.cloud import exceptions as google_cloud_exceptions
from opentelemetry import trace

logger = logging.getLogger(__name__)


class BigQueryExecutorError(exceptions.GarfExecutorError):
  """Error when BigQueryExecutor fails to run query."""


class BigQueryExecutor(executor.Executor, query_editor.TemplateProcessorMixin):
  """Handles query execution in BigQuery.

  Attributes:
      project_id: Google Cloud project id.
      location: BigQuery dataset location.
      client: BigQuery client.
  """

  def __init__(
    self,
    project_id: str | None = os.getenv('GOOGLE_CLOUD_PROJECT'),
    location: str | None = None,
    **kwargs: str,
  ) -> None:
    """Initializes BigQueryExecutor.

    Args:
        project_id: Google Cloud project id.
        location: BigQuery dataset location.
    """
    if not project_id:
      raise BigQueryExecutorError(
        'project_id is required. Either provide it as project_id parameter '
        'or GOOGLE_CLOUD_PROJECT env variable.'
      )
    self.project_id = project_id
    self.location = location
    super().__init__()

  @property
  def client(self) -> bigquery.Client:
    """Instantiates bigquery client."""
    return bigquery.Client(self.project_id)

  @tracer.start_as_current_span('bq.execute')
  def execute(
    self,
    query: str,
    title: str,
    context: execution_context.ExecutionContext = (
      execution_context.ExecutionContext()
    ),
  ) -> report.GarfReport:
    """Executes query in BigQuery.

    Args:
      query: Location of the query.
      title: Name of the query.
      context: Query execution context.

    Returns:
      Report with data if query returns some data otherwise empty Report.
    """
    span = trace.get_current_span()
    span.set_attribute('query.title', title)
    span.set_attribute('query.text', query)
    logger.info('Executing script: %s', title)
    query_text = self.replace_params_template(query, context.query_parameters)
    self.create_datasets(context.query_parameters.macro)
    job = self.client.query(query_text)
    try:
      result = job.result()
    except google_cloud_exceptions.GoogleCloudError as e:
      raise BigQueryExecutorError(
        f'Failed to execute query {title}: Reason: {e}'
      ) from e
      logger.debug('%s launched successfully', title)
    if result.total_rows:
      results = report.GarfReport.from_pandas(result.to_dataframe())
    else:
      results = report.GarfReport()
    if context.writer and results:
      writer_clients = context.writer_clients
      if not writer_clients:
        logger.warning('No writers configured, skipping write operation')
      else:
        writing_results = []
        for writer_client in writer_clients:
          logger.debug(
            'Start writing data for query %s via %s writer',
            title,
            type(writer_client),
          )
          writing_result = writer_client.write(results, title)
          logger.debug(
            'Finish writing data for query %s via %s writer',
            title,
            type(writer_client),
          )
          writing_results.append(writing_result)
        # Return the last writer's result for backward compatibility
        logger.info('%s executed successfully', title)
        return writing_results[-1] if writing_results else None
    logger.info('%s executed successfully', title)
    span.set_attribute('execute.num_results', len(results))
    return results

  @tracer.start_as_current_span('bq.create_datasets')
  def create_datasets(self, macros: dict | None) -> None:
    """Creates datasets in BQ based on values in a dict.

    If dict contains keys with 'dataset' in them, then values for such keys
    are treated as dataset names.

    Args:
      macros: Mapping containing data for query execution.
    """
    if macros and (datasets := extract_datasets(macros)):
      for dataset in datasets:
        dataset_id = f'{self.project_id}.{dataset}'
        try:
          self.client.get_dataset(dataset_id)
        except google_cloud_exceptions.NotFound:
          bq_dataset = bigquery.Dataset(dataset_id)
          bq_dataset.location = self.location
          with contextlib.suppress(google_cloud_exceptions.Conflict):
            self.client.create_dataset(bq_dataset, timeout=30)
            logger.info('Created new dataset %s', dataset_id)


def extract_datasets(macros: dict | None) -> list[str]:
  """Finds dataset-related keys based on values in a dict.

  If dict contains keys with 'dataset' in them, then values for such keys
  are treated as dataset names.

  Args:
      macros: Mapping containing data for query execution.

  Returns:
      Possible names of datasets.
  """
  if not macros:
    return []
  return [value for macro, value in macros.items() if 'dataset' in macro]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Stores mapping between API aliases and their execution context."""

from __future__ import annotations

import os
import pathlib

import pydantic
import smart_open
import yaml
from garf.executors.execution_context import ExecutionContext


class Config(pydantic.BaseModel):
  """Stores necessary parameters for one or multiple API sources.

  Attributes:
    source: Mapping between API source alias and execution parameters.
  """

  sources: dict[str, ExecutionContext]

  @classmethod
  def from_file(cls, path: str | pathlib.Path | os.PathLike[str]) -> Config:
    """Builds config from local or remote yaml file."""
    with smart_open.open(path, 'r', encoding='utf-8') as f:
      data = yaml.safe_load(f)
    return Config(sources=data)

  def save(self, path: str | pathlib.Path | os.PathLike[str]) -> str:
    """Saves config to local or remote yaml file."""
    with smart_open.open(path, 'w', encoding='utf-8') as f:
      yaml.dump(
        self.model_dump(exclude_none=True).get('sources'), f, encoding='utf-8'
      )
    return f'Config is saved to {str(path)}'



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""qQuery can be used as a parameter in garf queries."""

import contextlib

from garf.core import query_editor, query_parser
from garf.executors import execution_context


class GqueryError(query_parser.GarfQueryError):
  """Errors on incorrect qQuery syntax."""


def _handle_sub_context(context, sub_context):
  for k, v in sub_context.items():
    if isinstance(v, str) and v.startswith('gquery'):
      no_writer_context = context.model_copy(update={'writer': None})
      try:
        _, alias, *query = v.split(':', maxsplit=3)
      except ValueError:
        raise GqueryError(
          f'Incorrect gquery format, should be gquery:alias:query, got {v}'
        )
      if not alias:
        raise GqueryError(f'Missing alias in gquery: {v}')
      if not query:
        raise GqueryError(f'Missing query text in gquery: {v}')
      if alias == 'sqldb':
        from garf.executors import sql_executor

        gquery_executor = (
          sql_executor.SqlAlchemyQueryExecutor.from_connection_string(
            context.fetcher_parameters.get('connection_string')
          )
        )
      elif alias == 'bq':
        from garf.executors import bq_executor

        gquery_executor = bq_executor.BigQueryExecutor(
          **context.fetcher_parameters
        )
      else:
        raise GqueryError(f'Unsupported alias {alias} for gquery: {v}')
      with contextlib.suppress(
        query_editor.GarfResourceError, query_parser.GarfVirtualColumnError
      ):
        query = ':'.join(query)
        query_spec = query_editor.QuerySpecification(
          text=query, args=context.query_parameters
        ).generate()
        if len(columns := [c for c in query_spec.column_names if c != '_']) > 1:
          raise GqueryError(f'Multiple columns in gquery definition: {columns}')
      res = gquery_executor.execute(
        query=query, title='gquery', context=no_writer_context
      )
      if len(columns := [c for c in res.column_names if c != '_']) > 1:
        raise GqueryError(f'Multiple columns in gquery result: {columns}')
      sub_context[k] = res.to_list(row_type='scalar')


def process_gquery(
  context: execution_context.ExecutionContext,
) -> execution_context.ExecutionContext:
  _handle_sub_context(context, context.fetcher_parameters)
  _handle_sub_context(context, context.query_parameters.macro)
  return context



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for executing Garf queries and writing them to local/remote.

ApiQueryExecutor performs fetching data from API in a form of
GarfReport and saving it to local/remote storage.
"""
# pylint: disable=C0330, g-bad-import-order, g-multiple-import

from __future__ import annotations

import logging
import pathlib

from garf.core import report_fetcher, simulator
from garf.executors import (
  exceptions,
  execution_context,
  executor,
  fetchers,
  query_processor,
)
from garf.executors.telemetry import tracer
from opentelemetry import metrics, trace

logger = logging.getLogger(__name__)
meter = metrics.get_meter('garf.executors')

api_counter = meter.create_counter(
  'garf_api_execute_total',
  unit='1',
  description='Counts number of API executions',
)


class ApiExecutionContext(execution_context.ExecutionContext):
  """Common context for executing one or more queries."""

  writer: str | list[str] = 'console'


class ApiQueryExecutor(executor.Executor):
  """Gets data from API and writes them to local/remote storage.

  Attributes:
      api_client: a client used for connecting to API.
  """

  def __init__(
    self,
    fetcher: report_fetcher.ApiReportFetcher,
    report_simulator: simulator.ApiReportSimulator | None = None,
  ) -> None:
    """Initializes ApiQueryExecutor.

    Args:
      fetcher: Instantiated report fetcher.
      report_simulator: Instantiated simulator.
    """
    self.fetcher = fetcher
    self.simulator = report_simulator
    super().__init__(
      preprocessors=self.fetcher.preprocessors,
      postprocessors=self.fetcher.postprocessors,
    )

  @classmethod
  def from_fetcher_alias(
    cls,
    source: str,
    fetcher_parameters: dict[str, str] | None = None,
    enable_cache: bool = False,
    cache_ttl_seconds: int = 3600,
  ) -> ApiQueryExecutor:
    if not fetcher_parameters:
      fetcher_parameters = {}
    concrete_api_fetcher = fetchers.get_report_fetcher(source)
    return ApiQueryExecutor(
      fetcher=concrete_api_fetcher(
        **fetcher_parameters,
        enable_cache=enable_cache,
        cache_ttl_seconds=cache_ttl_seconds,
      )
    )

  @tracer.start_as_current_span('api.execute')
  def execute(
    self,
    query: str,
    title: str,
    context: ApiExecutionContext,
  ) -> str:
    """Reads query, extract results and stores them in a specified location.

    Args:
      query: Location of the query.
      title: Name of the query.
      context: Query execution context.

    Returns:
      Result of writing the report.

    Raises:
      GarfExecutorError: When failed to execute query.
    """
    if self.simulator:
      return self.simulate(query=query, title=title, context=context)
    context = query_processor.process_gquery(context)
    span = trace.get_current_span()
    span.set_attribute('fetcher.class', self.fetcher.__class__.__name__)
    span.set_attribute(
      'api.client.class', self.fetcher.api_client.__class__.__name__
    )
    try:
      span.set_attribute('query.title', title)
      span.set_attribute('query.text', query)
      logger.debug('starting query %s', query)
      title = pathlib.Path(title).name.split('.')[0]
      api_counter.add(
        1, {'api.client.class': self.fetcher.api_client.__class__.__name__}
      )
      results = self.fetcher.fetch(
        query_specification=query,
        args=context.query_parameters,
        title=title,
        **context.fetcher_parameters,
      )
      writer_clients = context.writer_clients
      if not writer_clients:
        logger.warning('No writers configured, skipping write operation')
        return None
      writing_results = []
      for writer_client in writer_clients:
        logger.debug(
          'Start writing data for query %s via %s writer',
          title,
          type(writer_client),
        )
        result = writer_client.write(results, title)
        logger.debug(
          'Finish writing data for query %s via %s writer',
          title,
          type(writer_client),
        )
        writing_results.append(result)
      logger.info('%s executed successfully', title)
      # Return the last writer's result for backward compatibility
      return writing_results[-1] if writing_results else None
    except Exception as e:
      logger.error('%s generated an exception: %s', title, str(e))
      raise exceptions.GarfExecutorError(
        '%s generated an exception: %s', title, str(e)
      ) from e

  @tracer.start_as_current_span('api.simulate')
  def simulate(
    self,
    query: str,
    title: str,
    context: ApiExecutionContext,
  ) -> str:
    """Reads query, simulates results and stores them in a specified location.

    Args:
      query: Location of the query.
      title: Name of the query.
      context: Query execution context.

    Returns:
      Result of writing the report.

    Raises:
      GarfExecutorError: When failed to execute query.
    """
    context = query_processor.process_gquery(context)
    span = trace.get_current_span()
    span.set_attribute('fetcher.class', self.fetcher.__class__.__name__)
    span.set_attribute(
      'api.client.class', self.fetcher.api_client.__class__.__name__
    )
    try:
      span.set_attribute('query.title', title)
      span.set_attribute('query.text', query)
      logger.debug('starting query %s', query)
      title = pathlib.Path(title).name.split('.')[0]
      results = self.simulator.simulate(
        query_specification=query,
        title=title,
        **context.fetcher_parameters,
      )
      writer_clients = context.writer_clients
      if not writer_clients:
        logger.warning('No writers configured, skipping write operation')
        return None
      writing_results = []
      for writer_client in writer_clients:
        logger.debug(
          'Start writing data for query %s via %s writer',
          title,
          type(writer_client),
        )
        result = writer_client.write(results, title)
        logger.debug(
          'Finish writing data for query %s via %s writer',
          title,
          type(writer_client),
        )
        writing_results.append(result)
      logger.info('%s executed successfully', title)
      # Return the last writer's result for backward compatibility
      return writing_results[-1] if writing_results else None
    except Exception as e:
      logger.error('%s generated an exception: %s', title, str(e))
      raise exceptions.GarfExecutorError(
        '%s generated an exception: %s', title, str(e)
      ) from e



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Runs garf workflow."""

from __future__ import annotations

import logging
import pathlib
import re
from typing import Final

import yaml
from garf.executors import exceptions, setup
from garf.executors.telemetry import tracer
from garf.executors.workflows import workflow
from garf.io import reader

logger = logging.getLogger(__name__)

_REMOTE_FILES_PATTERN: Final[str] = (
  '^(http|gs|s3|aruze|hdfs|webhdfs|ssh|scp|sftp)'
)
_SCRIPT_PATH = pathlib.Path(__file__).parent


class WorkflowRunner:
  """Runs garf workflow.

  Attributes:
    workflow: Workflow to execute.
    wf_parent: Optional location of a workflow file.
    parallel_threshold: Max allowed parallelism for the queries in the workflow.
  """

  def __init__(
    self,
    execution_workflow: workflow.Workflow,
    wf_parent: pathlib.Path | str,
    parallel_threshold: int = 10,
  ) -> None:
    """Initializes WorkflowRunner."""
    self.workflow = execution_workflow
    self.wf_parent = wf_parent
    self.parallel_threshold = parallel_threshold

  @classmethod
  def from_file(
    cls,
    workflow_file: str | pathlib.Path,
  ) -> WorkflowRunner:
    """Initialized Workflow runner from a local or remote file."""
    if isinstance(workflow_file, str):
      workflow_file = pathlib.Path(workflow_file)
    execution_workflow = workflow.Workflow.from_file(workflow_file)
    return cls(
      execution_workflow=execution_workflow, wf_parent=workflow_file.parent
    )

  def run(
    self,
    enable_cache: bool = False,
    cache_ttl_seconds: int = 3600,
    selected_aliases: list[str] | None = None,
    skipped_aliases: list[str] | None = None,
  ) -> list[str]:
    skipped_aliases = skipped_aliases or []
    selected_aliases = selected_aliases or []
    reader_client = reader.create_reader('file')
    execution_results = []
    logger.info('Starting Garf Workflow...')
    for i, step in enumerate(self.workflow.steps, 1):
      step_name = f'{i}-{step.fetcher}'
      if step.alias:
        step_name = f'{step_name}-{step.alias}'
      if step.alias in skipped_aliases:
        logger.warning(
          'Skipping step %d, fetcher: %s, alias: %s',
          i,
          step.fetcher,
          step.alias,
        )
        continue
      if selected_aliases and step.alias not in selected_aliases:
        logger.warning(
          'Skipping step %d, fetcher: %s, alias: %s',
          i,
          step.fetcher,
          step.alias,
        )
        continue
      with tracer.start_as_current_span(step_name):
        logger.info(
          'Running step %d, fetcher: %s, alias: %s', i, step.fetcher, step.alias
        )
        query_executor = setup.setup_executor(
          source=step.fetcher,
          fetcher_parameters=step.fetcher_parameters,
          enable_cache=enable_cache,
          cache_ttl_seconds=cache_ttl_seconds,
        )
        batch = {}
        if not (queries := step.queries):
          logger.error('Please provide one or more queries to run')
          raise exceptions.GarfExecutorError(
            'Please provide one or more queries to run'
          )
        for query in queries:
          if isinstance(query, workflow.QueryPath):
            query_path = query.full_path
            if re.match(_REMOTE_FILES_PATTERN, query_path):
              batch[query.path] = reader_client.read(query_path)
            else:
              if not query.prefix:
                query_path = self.wf_parent / pathlib.Path(query.path)
              if not query_path.exists():
                raise workflow.GarfWorkflowError(
                  f'Query: {query_path} not found'
                )
              batch[query.path] = reader_client.read(query_path)
          elif isinstance(query, workflow.QueryFolder):
            query_path = self.wf_parent / pathlib.Path(query.folder)
            if not query_path.exists():
              raise workflow.GarfWorkflowError(
                f'Folder: {query_path} not found'
              )
            for p in query_path.rglob('*'):
              if p.suffix == '.sql':
                batch[p.stem] = reader_client.read(p)
          else:
            batch[query.query.title] = query.query.text
        query_executor.execute_batch(
          batch, step.context, self.parallel_threshold
        )
        execution_results.append(step_name)
    return execution_results

  def compile(self, path: str | pathlib.Path) -> str:
    """Saves workflow with expanded anchors."""
    return self.workflow.save(path)

  def deploy(self, path: str | pathlib.Path) -> str:
    """Prepares workflow for deployment to Google Cloud Workflows."""
    wf = self.workflow.model_dump(exclude_none=True).get('steps')
    with open(_SCRIPT_PATH / 'gcp_workflow.yaml', 'r', encoding='utf-8') as f:
      cloud_workflow_run_template = yaml.safe_load(f)
    init = {
      'init': {
        'assign': [{'pairs': wf}],
      },
    }
    cloud_workflow = {
      'main': {
        'params': [],
        'steps': [init, cloud_workflow_run_template],
      },
    }
    with open(path, 'w', encoding='utf-8') as f:
      yaml.dump(cloud_workflow, f, sort_keys=False)
    return f'Workflow is saved to {path}'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gRPC endpoint for garf."""

import argparse
import logging
from concurrent import futures

import grpc
from garf.executors import execution_context, garf_pb2, garf_pb2_grpc, setup
from garf.executors.entrypoints.tracer import initialize_tracer
from google.protobuf.json_format import MessageToDict
from grpc_reflection.v1alpha import reflection


class GarfService(garf_pb2_grpc.GarfService):
  def Execute(self, request, context):
    query_executor = setup.setup_executor(
      request.source, request.context.fetcher_parameters
    )
    result = query_executor.execute(
      query=request.query,
      title=request.title,
      context=execution_context.ExecutionContext(
        **MessageToDict(request.context, preserving_proto_field_name=True)
      ),
    )
    return garf_pb2.ExecuteResponse(results=[result])

  def Fetch(self, request, context):
    query_executor = setup.setup_executor(
      request.source, request.context.fetcher_parameters
    )
    query_args = execution_context.ExecutionContext(
      **MessageToDict(request.context, preserving_proto_field_name=True)
    ).query_parameters
    result = query_executor.fetcher.fetch(
      query_specification=request.query,
      title=request.title,
      args=query_args,
    )
    return garf_pb2.FetchResponse(
      columns=result.column_names, rows=result.to_list(row_type='dict')
    )


if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument('--port', dest='port', default=50051, type=int)
  parser.add_argument(
    '--parallel-threshold', dest='parallel_threshold', default=10, type=int
  )
  args, _ = parser.parse_known_args()
  initialize_tracer()
  server = grpc.server(
    futures.ThreadPoolExecutor(max_workers=args.parallel_threshold)
  )

  service = GarfService()
  garf_pb2_grpc.add_GarfServiceServicer_to_server(service, server)
  SERVICE_NAMES = (
    garf_pb2.DESCRIPTOR.services_by_name['GarfService'].full_name,
    reflection.SERVICE_NAME,
  )
  reflection.enable_server_reflection(SERVICE_NAMES, server)
  server.add_insecure_port(f'[::]:{args.port}')
  server.start()
  logging.info('Garf service started, listening on port %d', 50051)
  server.wait_for_termination()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import pathlib
import re

import pydantic
import smart_open
import yaml
from garf.executors import exceptions
from garf.executors.execution_context import ExecutionContext


class GarfWorkflowError(exceptions.GarfExecutorError):
  """Workflow specific exception."""


class QueryFolder(pydantic.BaseModel):
  """Path to folder with queries."""

  folder: str


class QueryPath(pydantic.BaseModel):
  """Path file with query."""

  path: str
  prefix: str | None = None

  @property
  def full_path(self) -> str:
    if self.prefix:
      return re.sub('/$', '', self.prefix) + '/' + self.path
    return self.path


class QueryDefinition(pydantic.BaseModel):
  """Definition of a query."""

  query: Query


class Query(pydantic.BaseModel):
  """Query elements.

  Attributes:
    text: Query text.
    title: Name of the query.
  """

  text: str
  title: str


class ExecutionStep(ExecutionContext):
  """Common context for executing one or more queries.

  Attributes:
    fetcher: Name of a fetcher to get data from API.
    alias: Optional alias to identify execution step.
    queries: Queries to run for a particular fetcher.
    context: Execution context for queries and fetcher.
  """

  fetcher: str | None = None
  alias: str | None = pydantic.Field(default=None, pattern=r'^[a-zA-Z0-9_]+$')
  queries: list[QueryPath | QueryDefinition | QueryFolder] | None = None

  @property
  def context(self) -> ExecutionContext:
    return ExecutionContext(
      writer=self.writer,
      writer_parameters=self.writer_parameters,
      query_parameters=self.query_parameters,
      fetcher_parameters=self.fetcher_parameters,
    )


class Workflow(pydantic.BaseModel):
  """Orchestrates execution of queries for multiple fetchers.

  Attributes:
    steps: Contains one or several fetcher executions.
  """

  steps: list[ExecutionStep]

  @classmethod
  def from_file(cls, path: str | pathlib.Path | os.PathLike[str]) -> Workflow:
    """Builds workflow from local or remote yaml file."""
    with smart_open.open(path, 'r', encoding='utf-8') as f:
      data = yaml.safe_load(f)
    try:
      return Workflow(**data)
    except pydantic.ValidationError as e:
      raise GarfWorkflowError(f'Incorrect workflow:\n {e}') from e

  def save(self, path: str | pathlib.Path | os.PathLike[str]) -> str:
    """Saves workflow to local or remote yaml file."""
    with smart_open.open(path, 'w', encoding='utf-8') as f:
      yaml.dump(
        self.model_dump(exclude_none=True).get('steps'), f, encoding='utf-8'
      )
    return f'Workflow is saved to {str(path)}'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining `garf` CLI utility.

`garf` allows to execute queries and store results in local/remote
storage.
"""

from __future__ import annotations

import argparse
import logging
import pathlib
import sys

import garf.executors
from garf.executors import config, exceptions, setup
from garf.executors.entrypoints import utils
from garf.executors.entrypoints.tracer import (
  initialize_meter,
  initialize_tracer,
)
from garf.executors.telemetry import tracer
from garf.executors.workflows import workflow, workflow_runner
from garf.io import reader
from opentelemetry import trace

initialize_tracer()
meter_provider = initialize_meter()


@tracer.start_as_current_span('garf.entrypoints.cli')
def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('query', nargs='*')
  parser.add_argument('-c', '--config', dest='config', default=None)
  parser.add_argument('-w', '--workflow', dest='workflow', default=None)
  parser.add_argument('--source', dest='source', default=None)
  parser.add_argument('--output', dest='output', default='console')
  parser.add_argument('--input', dest='input', default='file')
  parser.add_argument('--log', '--loglevel', dest='loglevel', default='info')
  parser.add_argument('--logger', dest='logger', default='local')
  parser.add_argument('--log-name', dest='log_name', default='garf')
  parser.add_argument(
    '--parallel-queries', dest='parallel_queries', action='store_true'
  )
  parser.add_argument(
    '--no-parallel-queries', dest='parallel_queries', action='store_false'
  )
  parser.add_argument('--simulate', dest='simulate', action='store_true')
  parser.add_argument('--dry-run', dest='dry_run', action='store_true')
  parser.add_argument('-v', '--version', dest='version', action='store_true')
  parser.add_argument(
    '--parallel-threshold', dest='parallel_threshold', default=10, type=int
  )
  parser.add_argument(
    '--enable-cache', dest='enable_cache', action='store_true'
  )
  parser.add_argument(
    '--cache-ttl-seconds',
    dest='cache_ttl_seconds',
    default=3600,
    type=int,
  )
  parser.add_argument('--workflow-skip', dest='workflow_skip', default=None)
  parser.add_argument(
    '--workflow-include', dest='workflow_include', default=None
  )
  parser.set_defaults(parallel_queries=True)
  parser.set_defaults(simulate=False)
  parser.set_defaults(enable_cache=False)
  parser.set_defaults(dry_run=False)
  args, kwargs = parser.parse_known_args()

  span = trace.get_current_span()
  command_args = ' '.join(sys.argv[1:])
  span.set_attribute('cli.command', f'garf {command_args}')
  if args.version:
    print(garf.executors.__version__)
    sys.exit()
  logger = utils.init_logging(
    loglevel=args.loglevel.upper(), logger_type=args.logger, name=args.log_name
  )
  reader_client = reader.create_reader(args.input)
  if workflow_file := args.workflow:
    wf_parent = pathlib.Path.cwd() / pathlib.Path(workflow_file).parent
    execution_workflow = workflow.Workflow.from_file(workflow_file)
    workflow_skip = args.workflow_skip if args.workflow_skip else None
    workflow_include = args.workflow_include if args.workflow_include else None
    workflow_runner.WorkflowRunner(
      execution_workflow=execution_workflow, wf_parent=wf_parent
    ).run(
      enable_cache=args.enable_cache,
      cache_ttl_seconds=args.cache_ttl_seconds,
      selected_aliases=workflow_include,
      skipped_aliases=workflow_skip,
    )
    meter_provider.shutdown()
    sys.exit()

  if not args.query:
    logger.error('Please provide one or more queries to run')
    raise exceptions.GarfExecutorError(
      'Please provide one or more queries to run'
    )
  if config_file := args.config:
    execution_config = config.Config.from_file(config_file)
    if not (context := execution_config.sources.get(args.source)):
      raise exceptions.GarfExecutorError(
        f'No execution context found for source {args.source} in {config_file}'
      )
  else:
    param_types = ['source', 'macro', 'template']
    outputs = args.output.split(',')
    extra_parameters = utils.ParamsParser([*param_types, *outputs]).parse(
      kwargs
    )
    source_parameters = extra_parameters.get('source', {})
    writer_parameters = {}
    for output in outputs:
      writer_parameters.update(extra_parameters.get(output))

    context = garf.executors.api_executor.ApiExecutionContext(
      query_parameters={
        'macro': extra_parameters.get('macro'),
        'template': extra_parameters.get('template'),
      },
      writer=outputs,
      writer_parameters=writer_parameters,
      fetcher_parameters=source_parameters,
    )
  query_executor = setup.setup_executor(
    source=args.source,
    fetcher_parameters=context.fetcher_parameters,
    enable_cache=args.enable_cache,
    cache_ttl_seconds=args.cache_ttl_seconds,
    simulate=args.simulate,
  )
  batch = {query: reader_client.read(query) for query in args.query}
  if args.parallel_queries and len(args.query) > 1:
    logger.info('Running queries in parallel')
    batch = {query: reader_client.read(query) for query in args.query}
    query_executor.execute_batch(batch, context, args.parallel_threshold)
  else:
    if len(args.query) > 1:
      logger.info('Running queries sequentially')
    for query in args.query:
      query_executor.execute(
        query=reader_client.read(query), title=query, context=context
      )
  logging.shutdown()
  meter_provider.shutdown()


if __name__ == '__main__':
  main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI endpoint for executing queries."""

from typing import Optional, Union

import fastapi
import garf.executors
import pydantic
import typer
import uvicorn
from garf.executors import exceptions, setup
from garf.executors.entrypoints import utils
from garf.executors.entrypoints.tracer import (
  initialize_meter,
  initialize_tracer,
)
from garf.executors.workflows import workflow_runner
from garf.io import reader
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Annotated

initialize_tracer()
initialize_meter()
app = fastapi.FastAPI()
FastAPIInstrumentor.instrument_app(app)
typer_app = typer.Typer()


class GarfSettings(BaseSettings):
  """Specifies environmental variables for garf.

  Ensure that mandatory variables are exposed via
  export ENV_VARIABLE_NAME=VALUE.

  Attributes:
    loglevel: Level of logging.
    log_name: Name of log.
    logger_type: Type of logger.
  """

  model_config = SettingsConfigDict(env_prefix='garf_')

  loglevel: str = 'INFO'
  log_name: str = 'garf'
  logger_type: str = 'local'


class GarfDependencies:
  def __init__(self) -> None:
    """Initializes GarfDependencies."""
    settings = GarfSettings()
    self.logger = utils.init_logging(
      loglevel=settings.loglevel,
      logger_type=settings.logger_type,
      name=settings.log_name,
    )


class ApiExecutorRequest(pydantic.BaseModel):
  """Request for executing a query.

  Attributes:
    source: Type of API to interact with.
    title: Name of the query used as an output for writing.
    query: Query to execute.
    query_path: Local or remote path to query.
    context: Execution context.
  """

  source: str
  title: Optional[str] = None
  query: Optional[str] = None
  query_path: Optional[Union[str, list[str]]] = None
  context: garf.executors.api_executor.ApiExecutionContext

  @pydantic.model_validator(mode='after')
  def check_query_specified(self):
    if not self.query_path and not self.query:
      raise exceptions.GarfExecutorError(
        'Missing one of required parameters: query, query_path'
      )
    return self

  def model_post_init(self, __context__) -> None:
    if self.query_path and isinstance(self.query_path, str):
      self.query = reader.FileReader().read(self.query_path)
    if not self.title:
      self.title = str(self.query_path)


class ApiExecutorResponse(pydantic.BaseModel):
  """Response after executing a query.

  Attributes:
    results: Results of query execution.
  """

  results: list[str]


@app.get('/api/version')
async def version() -> str:
  return garf.executors.__version__


@app.get('/api/fetchers')
async def get_fetchers(
  dependencies: Annotated[GarfDependencies, fastapi.Depends(GarfDependencies)],
) -> list[str]:
  """Shows all available API sources."""
  return list(garf.executors.fetchers.find_fetchers())


@app.post('/api/execute')
def execute(
  request: ApiExecutorRequest,
  dependencies: Annotated[GarfDependencies, fastapi.Depends(GarfDependencies)],
) -> ApiExecutorResponse:
  query_executor = setup.setup_executor(
    request.source, request.context.fetcher_parameters
  )
  result = query_executor.execute(request.query, request.title, request.context)
  return ApiExecutorResponse(results=[result])


@app.post('/api/execute:batch')
def execute_batch(
  request: ApiExecutorRequest,
  dependencies: Annotated[GarfDependencies, fastapi.Depends(GarfDependencies)],
) -> ApiExecutorResponse:
  query_executor = setup.setup_executor(
    request.source, request.context.fetcher_parameters
  )
  reader_client = reader.FileReader()
  batch = {query: reader_client.read(query) for query in request.query_path}
  results = query_executor.execute_batch(batch, request.context)
  return ApiExecutorResponse(results=results)


@app.post('/api/execute:workflow')
def execute_workflow(
  workflow_file: str,
  dependencies: Annotated[GarfDependencies, fastapi.Depends(GarfDependencies)],
  enable_cache: bool = False,
  cache_ttl_seconds: int = 3600,
) -> list[str]:
  return workflow_runner.WorkflowRunner.from_file(workflow_file).run(
    enable_cache=enable_cache, cache_ttl_seconds=cache_ttl_seconds
  )


@typer_app.command()
def main(
  port: Annotated[int, typer.Option(help='Port to start the server')] = 8000,
):
  uvicorn.run(app, port=port)


if __name__ == '__main__':
  typer_app()
